#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <wx/wx.h>
#include <wx/scrolbar.h>
#include <wx/print.h>
#include <wx/printdlg.h>
#include <wx/accel.h>
#include <wx/colordlg.h>

#include "afhelp.h"
#include "afwx.h"
#include "afview.h"
#include "crect.h"
#include "painter.h"


void ViewInfo::WriteToConfig(wxConfig *cfg)
{
	cfg->Write("psz", this->psz);
	cfg->Write("showgrid", this->showgrid);
}

void ViewInfo::ReadFromConfig(wxConfig *cfg)
{
	cfg->Read("psz", &this->psz);
	cfg->Read("showgrid", &this->showgrid);
}

BEGIN_EVENT_TABLE( AFView, wxWindow )
	EVT_LEFT_DOWN( AFView::OnButtonDown )
	EVT_LEFT_UP( AFView::OnButtonUp )

	EVT_RIGHT_DOWN( AFView::OnButtonDown )
	EVT_RIGHT_UP( AFView::OnButtonUp )

	EVT_MOTION( AFView::OnMouseMove )
	EVT_CHAR( AFView::OnKeyEvent )
	EVT_PAINT( AFView::OnPaint )
	EVT_SIZE( AFView::OnSize )
END_EVENT_TABLE()


AFView::AFView(AFFrame *frm)
	: wxWindow(frm, -1, wxPoint(10, 10), wxSize(200, 200), wxWANTS_CHARS)
{
	/* setup initial pointers */
	mFrame = frm;
	bIsPrinting = false;
	
	/* allocate the double buffer bitmap */
	mBuffer = new wxBitmap(300, 400);


	mSelectColor = *wxRED;
	mForeground = *wxBLACK;
	mBackground = *wxWHITE;
	mMaskColor = wxColour(255,0,255);
	
	mCursorDefault = wxCursor(wxCURSOR_PENCIL);
	mCursorDelete = wxCursor(wxCURSOR_NO_ENTRY);
	mCursorCopy = wxCursor(wxCURSOR_HAND);
	mCursorInsert = wxCursor(wxCURSOR_POINT_LEFT);

	SetBackgroundColour( mBackground );
	SetCursor( mCursorDefault );
	
	/* initialize view properties parameters */
	mVI.psz = 10;
	mVI.xoff = 5;
	mVI.yoff = 30;
	mVI.showgrid = true;
	mVI.ReadFromConfig( app_config );

	bDirty = false;

	mColor1 = *wxBLUE;
	mColor2 = *wxLIGHT_GREY;

	wxSize geom = GetSize();

	mVI.pixwidth = geom.GetWidth();
	mVI.pixheight = geom.GetHeight();

	mClientDC = new wxClientDC(this);
	mClientPainter = new Painter( mClientDC );


	mImage.SetMask(true);
	mImage.SetMaskColour( mMaskColor.Red(), mMaskColor.Green(), mMaskColor.Blue() );

}

AFView::~AFView()
{
	mVI.WriteToConfig( app_config );

	delete mBuffer;

	delete mClientPainter;
	delete mClientDC;
}

void AFView::ZoomIn()
{
	mVI.psz += 1;
	if (mVI.psz > 20)
		mVI.psz = 20;

	RedrawAll();
}

void AFView::ZoomOut()
{
	mVI.psz -= 1;
	if (mVI.psz < 1)
		mVI.psz = 1;

	RedrawAll();
}


void AFView::GetViewInfo(ViewInfo *info)
{
	if (!info)
		return;

	*info = this->mVI;
}

void AFView::SetViewInfo(ViewInfo &info, bool redraw)
{
	this->mVI = info;

	if (redraw)
		RedrawAll();
}

void AFView::SetPrinting( bool b )
{
	bIsPrinting = b;
}

bool AFView::IsPrinting()
{
	return bIsPrinting;
}


wxImage &AFView::GetImage()
{
	return mImage;
}
	
void AFView::RedrawAll()
{
	wxMemoryDC dc;
	dc.SelectObject(*mBuffer);
	
	
	Draw(dc);

	wxClientDC cdc(this);
	cdc.Blit(0, 0, mBuffer->GetWidth(), mBuffer->GetHeight(), &dc, 0, 0);

	dc.SelectObject( wxNullBitmap );
}

bool AFView::LoadFile(const wxString &name, int type)
{

	if (mImage.LoadFile( name , type ))
	{
		if (mImage.HasAlpha())
			wxMessageBox("Image has alpha support!");
		bDirty = false;
		mImage.SetMask(true);
		mImage.SetMaskColour( mMaskColor.Red(), mMaskColor.Green(), mMaskColor.Blue() );
		RedrawAll();
		return true;
	}
	else
		return false;
}

bool AFView::SaveFile(const wxString &name, int type)
{
	mImage.SetMask(true);
	mImage.SetMaskColour( mMaskColor.Red(), mMaskColor.Green(), mMaskColor.Blue() );

	if (mImage.SaveFile( name, type ))
	{
		bDirty = false;
		return true;
	}
	else
		return false;
}

void AFView::Clear(int w, int h)
{
	mImage = wxImage( w, h );
		
	unsigned char *data = mImage.GetData();
	int len = w * h * 3;
	for (int i=0;i<len;i++)
		data[i] = 255;
}

void AFView::Resize(const wxSize &sz)
{
	if (sz.GetWidth() > 0 && sz.GetHeight() > 0)
	{
		mImage.SetMaskColour( mMaskColor.Red(), mMaskColor.Green(), mMaskColor.Blue() );
		mImage.Resize(sz, wxPoint(0,0));
		RedrawAll();
	}
}

void AFView::Draw(wxDC &dc)
{

	Painter gfx(&dc);
	gfx.SetColour(*wxWHITE);
	gfx.SetFill(Painter::SOLID);
	gfx.DrawRectangle(0, 0, mVI.pixwidth, mVI.pixheight);
	gfx.SetCourierMedium();
	
	if (mImage.GetWidth() == 0 || mImage.GetHeight() == 0)
	{
		gfx.DrawString(10, 10, "No XPM image loaded.");
	}
	else
	{
		int width, height, i, x, y, idx;
		width = mImage.GetWidth();
		height = mImage.GetHeight();

		gfx.SetFill(Painter::SOLID);
		
		gfx.SetColour( mColor1 );
		gfx.DrawRectangle(mVI.xoff, 5, 15, 15);
		gfx.SetColour( mColor2 );
		gfx.DrawRectangle(mVI.xoff+20, 5, 15, 15);
		gfx.SetColour( mMaskColor );
		gfx.DrawRectangle(mVI.xoff+40, 5, 15, 15);

		gfx.SetColour(*wxBLACK);
		gfx.SetFill(Painter::EMPTY);
		gfx.DrawRectangle(mVI.xoff, 5, 15, 15);
		gfx.DrawRectangle(mVI.xoff+20, 5, 15, 15);

		gfx.DrawString(mVI.xoff+60, 5, wxString::Format("Image %dx%d Zoom %dx", width, height, mVI.psz));
		

		unsigned char *data = mImage.GetData();

		int bpp = 3;
		gfx.SetFill(Painter::SOLID);

		for (y=0; y < height; y++)
		{
			for (x=0; x < width; x++)
			{		
				idx = (y * width) + x;
				gfx.SetColour( wxColour( data[idx*bpp], data[idx*bpp+1], data[idx*bpp+2] ));

				if (mVI.psz < 2)
					gfx.DrawPoint( mVI.xoff + x*mVI.psz, mVI.yoff+y*mVI.psz );
				else
					gfx.DrawRectangle(mVI.xoff+x*mVI.psz, mVI.yoff+y*mVI.psz, mVI.psz, mVI.psz);
			}
		}

		DrawGrid(gfx);
	}
}

void AFView::DrawGrid(Painter &gfx)
{
	if (mVI.showgrid && mVI.psz > 2)
	{
		int width, height, i;
		width = mImage.GetWidth();
		height = mImage.GetHeight();

		gfx.SetColour(*wxBLACK);
		gfx.SetFill(Painter::EMPTY);
		
		gfx.DrawRectangle(mVI.xoff, mVI.yoff, width * mVI.psz+1, height * mVI.psz+1);

		// horizontal lines
		for (i=0;i<width;i++)
			gfx.DrawLine(mVI.xoff, mVI.yoff + i*mVI.psz, mVI.xoff+width*mVI.psz, mVI.yoff + i*mVI.psz);

		// vertical lines
		for (i=0;i<height;i++)
			gfx.DrawLine(mVI.xoff+i*mVI.psz, mVI.yoff, mVI.xoff+i*mVI.psz,mVI.yoff + height*mVI.psz);
	}
}

void AFView::DrawPix(int cx, int cy, const wxColour &c)
{
	if (cx >=0 && cy >= 0 && cx < mImage.GetWidth() && cy < mImage.GetHeight())
	{
		mImage.SetRGB(cx, cy, c.Red(), c.Green(), c.Blue());

		mClientPainter->SetColour( c );
		mClientPainter->SetFill(Painter::SOLID);
		mClientPainter->DrawRectangle(mVI.xoff+cx*mVI.psz, mVI.yoff+cy*mVI.psz, mVI.psz, mVI.psz);

		DrawGrid( *mClientPainter );

		bDirty = true;
		mFrame->UpdateControlStatus();
		mFrame->UpdateTitlebar(false);
	}
}

void AFView::OnButtonDown(wxMouseEvent &evt)
{
	int x, y;
	x = evt.GetX();
	y = evt.GetY();

	if (mImage.GetWidth() > 0 && mImage.GetHeight() > 0)
	{
		int cx = (x - mVI.xoff)/mVI.psz;
		int cy = (y - mVI.yoff)/mVI.psz;

		wxColour c = evt.GetButton() == wxMOUSE_BTN_LEFT ? mColor1 : mColor2;
		DrawPix(cx, cy, c);
	}
}

void AFView::OnButtonUp(wxMouseEvent &evt)
{
}

void AFView::OnMouseMove(wxMouseEvent &evt)
{
	int x, y;
	x = evt.GetX();
	y = evt.GetY();

	if ((evt.LeftIsDown() || evt.RightIsDown() ) && mImage.GetWidth() > 0 && mImage.GetHeight() > 0)
	{
		int cx = (x - mVI.xoff)/mVI.psz;
		int cy = (y - mVI.yoff)/mVI.psz;

		wxColour c = evt.LeftIsDown() ? mColor1 : mColor2;
		DrawPix(cx, cy, c);
	}
}

void AFView::ChangeColor(wxColour *color)
{
	wxColour init = *color;
	wxColour c = wxGetColourFromUser(this, init);
	if (c.Ok())
	{
		*color = c;
		RedrawAll();
	}
}

void AFView::OnKeyEvent(wxKeyEvent &evt)
{
	bool handled = false;
	switch(evt.GetKeyCode())
	{
	case WXK_UP:
		handled = true;
		break;
	case WXK_DOWN:
		handled = true;
		break;
	case WXK_RIGHT:
		handled = true;
		break;
	case WXK_LEFT:
		handled = true;
		break;
	case WXK_ESCAPE:
		handled = true;
		break;
	case 'i':
	case 'I':
		ZoomIn();
		handled = true;
		break;
	case 'o':
	case 'O':
		ZoomOut();
		handled = true;
		break;
	case 'g':
	case 'G':
		mVI.showgrid = ! mVI.showgrid;
		RedrawAll();
		handled = true;
		break;
	case 't':
	case 'T':
		mColor1 = mMaskColor;
		RedrawAll();
		handled = true;
		break;
	case 'c':
	case 'C':
		ChangeColor(&mColor1);
		handled = true;
		break;
	case 'v':
	case 'V':
		ChangeColor(&mColor2);
		handled = true;
		break;
		break;
	}

	if (!handled)
		evt.Skip();
}

void AFView::OnSize(wxSizeEvent &evt)
{
	wxSize geom = evt.GetSize();
	mVI.pixwidth = geom.GetWidth();
	mVI.pixheight = geom.GetHeight();
	if (mBuffer->GetWidth() < geom.GetWidth() ||
		mBuffer->GetHeight() < geom.GetHeight() )
	{
		delete mBuffer;
		mBuffer = new wxBitmap(geom.GetWidth(), geom.GetHeight());
		RedrawAll();
	}
}

void AFView::OnPaint(wxPaintEvent &evt)
{
	/* should this just copy the double buffer rather ? */
	wxPaintDC dc(this);
	PrepareDC(dc);

	dc.BeginDrawing();
	Draw( dc );
	dc.EndDrawing();
}

		
