#include "tokenizer.h"

/*@
 c Tokenizer
 t The Tokenizer class provides a simple way to split
 t delimited strings into tokens.  The tokenizer can be
 t configured to recognize any set of single characters
 t as delimiters.  A tokenizer can be used either with
 t the <i>hasTokens()</i> and <i>next()</i> methods, or with
 t the more flexible <i>get()</i> and <i>unget()</i> methods.
 t When <i>get(..)</i> is called, it returns the next token
 t in a queue of tokens.  If the token is not ready to be accepted,
 t it can be returned to the queue in the correct order with the
 t <i>unget()</i> method, which can be called as many times as required.
 t This way, the list of tokenized items can be traversed back and forth
 t if necessary.
@*/

/*@
 f Tokenizer
 d Creates an empty space-delimited tokenizer.
@*/
Tokenizer::Tokenizer()
{
	head = NULL;
	tail = NULL;
	cur = NULL;

	str = "";
	pos = 0;
	delims = " ";
	last_pos = 0;
}

/*@
 f Tokenizer
 d Creates a new space-delimited tokenizer with the given input wxString.
 p const wxString& : s The wxString to tokenize.
@*/
Tokenizer::Tokenizer(const wxString &s)
{
	head = NULL;
	tail = NULL;
	cur = NULL;

	str = s;
	pos = 0;
	last_pos = 0;
	delims = " ";
}

/*@
 f Tokenizer
 d Creates a new tokenizer on the given input wxString with custom delimiters.
 p const wxString& : s The wxString to tokenize.
 p const wxString& : dlms A wxString of delimiter characters.
@*/
Tokenizer::Tokenizer(const wxString &s, const wxString &dlms)
{
	head = NULL;
	tail = NULL;
	cur = NULL;

	str = s;
	pos = 0;
	last_pos = 0;
	delims = dlms;
}

Tokenizer::~Tokenizer()
{
	clearQueue();
}

/*@
 f next
 d Returns the next token in the input. Use in conjunction with <i>hasTokens()</i>.
 r wxString
@*/
wxString Tokenizer::next()
{
	wxString token = "";
	int len = str.length();
	
	while (pos < len && delims.find(str[pos]) != -1)
		pos++;
	
	while (pos < len && delims.find(str[pos]) == -1)
		token += str[pos++];
	
	last_pos = pos;
	
	while (pos < len && delims.find(str[pos]) != -1)
		pos++;
	
	return token;
}

/*@
 f isEmpty
 d Determines whether the token queue is empty.
 r bool :
@*/
bool Tokenizer::isEmpty()
{
	if (head == NULL)
	{
		pump();
		cur = head; // reset cur to the front of the list
	}
	
	if (cur == NULL)
		return true;
	else
		return false;
}

/*@
 f get
 d Gets the next token in the queue.
 r bool : Returns false if the queue was empty.
 p wxString& : pTok The wxString to assign the next token to.
@*/
bool Tokenizer::get(wxString &tok)
{
	if (isEmpty())
		return false;
			
	if (cur->next == NULL)
		pump();
	
	tok = cur->str;
	cur = cur->next;
	
	return true;
}

/*@
 f unget
 d Ungets one token and puts it back on the front of the token queue.
 r bool : Returns false if a token could not be put back on the queue.
@*/
bool Tokenizer::unget()
{
	if (cur == NULL)
		cur = tail;
	
	if (cur == NULL)
		return false;
	
	if (cur->prev)
	{
		cur = cur->prev;
		return true;
	}
	
	return false;
}

/*@
 f clearQueue
 d Clears the token queue.
 r void :
@*/
void Tokenizer::clearQueue()
{
	while (head != NULL)
	{
		cur = head->next;
		delete head;
		head = cur;
	}
	
	head = NULL;
	tail = NULL;
	cur = NULL;
}

void Tokenizer::pump()
{
	if (hasTokens())
	{
		Token *tok = new Token;
		tok->str = next();
				
		if (tail == NULL)
		{
			tok->next = head;
			tok->prev = NULL;
			head = tok;
		}
		else
		{
			tail->next = tok;
			tok->next = NULL;
			tok->prev = tail;
		}
		
		tail = tok;
	}		
}


/*@
 f hasTokens
 d Returns whether tokens remain on the input wxString. Use with <i>next()</i>.
 r bool :
@*/
bool Tokenizer::hasTokens()
{
	return pos < (int)str.length();
}

/*@
 f getLastTokenEndPos
 d Returns the ending position of the most recently processed token in the input wxString.
 r int :
@*/
int Tokenizer::getLastTokenEndPos()
{
	return last_pos;
}

/*@
 f getCurrentPos
 d Returns the current position of the tokenizer in the input wxString.
 r int :
@*/
int Tokenizer::getCurrentPos()
{
	return pos;
}

/*@
 f getRestOfInput
 d Returns the rest of the unprocessed input wxString.
 r wxString :
@*/
wxString Tokenizer::getRestOfInput()
{
	return str.Mid(pos);
}

/*@
 f reset
 d Resets the tokenizer to a new input wxString. Note that this does not clear the token queue.
 r void :
 p const wxString& : s The new input wxString.
@*/
void Tokenizer::reset(const wxString &s)
{
	clearQueue();
	
	pos = 0;
	last_pos = 0;
	str = s;
}

/*@
 f setDelimiters
 d Changes the character delimiters.
 r void :
 p const wxString& : dlms A wxString of single character delimiters.
@*/
void Tokenizer::setDelimiters(const wxString &dlms)
{
	delims = dlms;
}

/*@
 f getDelimiters
 d Returns the current delimiters.
 r wxString
@*/
wxString Tokenizer::getDelimiters()
{
	return delims;
}

